#include <avr/io.h>
#include <avr/interrupt.h>
#include <EEPROM.h>
#include <LowPower.h>
#include <LoRa.h>


/* --------------------- */
#define LOGDEBUG  1
/* --------------------- */
#define LORA_SS    10 //PB2
#define LORA_RST   8  //PB0
#define LORA_DIO0  9  //PB1
/* --------------------- */
#define LORA_CH   0x17
/* --------------------- */
#define PULSE_OUT_A  3  //PD3
#define PULSE_OUT_B  4  //PD4
/* --------------------- */
#define PULSE_IN_A  15  //PC1 (ADC1)
#define PULSE_IN_B  14  //PC0 (ADC0)
/* --------------------- */
#define AIN0        6   //PD6
#define EN_VREF     5   //PD5
/* --------------------- */
#define LED         19   //PC5
/* --------------------- */
#define B_ADDRESS   0x10
#define B_SIZE      192
/* --------------------- */
#define SEND_PERIOD 300


uint8_t currentBufferPosition = 0;

volatile uint8_t pulseCntInt = 0;

uint8_t pulseCntA = 0;
uint8_t pulseCntAMax = 45;
uint8_t pulseCntAMin = 0;

uint8_t pulseCntB = 0;
uint8_t pulseCntBMax = 45;
uint8_t pulseCntBMin = 0;

uint8_t thresholdA = 0;
uint8_t thresholdB = 0;

bool calibrationRequested = false;
bool slowMode = false;


// State is four bits : [A, B, new A, new B], 0 without metal, 1 with metal
uint8_t rotationState = 0b0011;
int32_t rotationIndex = 0;
uint32_t errorIndex = 0;

//timers
uint32_t lastChangeTime = 0;
uint32_t lastTransferTime = 0;

uint8_t     plainRadioBuffer[25]        = {0};
uint8_t     transfertRadioBuffer[31]   = {0};

uint16_t measureCounter = 0;



bool updateState(){
  bool isChanged = true;
  
  //make state as old
  rotationState = (rotationState << 2) & 0b1111;

  //get disc position
  if(pulseCntA <= thresholdA) rotationState |= 0b0010;
  if(pulseCntB <= thresholdB) rotationState |= 0b0001;

  //treat transitions with events
  switch(rotationState){
    case 0b0000:
    case 0b0101:
    case 0b1010:
    case 0b1111:
      isChanged = false;
      break;
    
    case 0b0001:
      rotationIndex--;
      break;

    case 0b0100:
      rotationIndex++;
      updateEEPROM();
      break;  
        
    case 0b0011:
    case 0b0110:
    case 0b1001:
    case 0b1100:
      errorIndex++;
      break;  
  }

  return isChanged;
}

void setup() {
  Serial.begin(250000); 

  pinMode(PULSE_OUT_A , OUTPUT);
  digitalWrite(PULSE_OUT_A, LOW); 

  pinMode(PULSE_OUT_B , OUTPUT);
  digitalWrite(PULSE_OUT_B, LOW); 

  // Pulse in as output with low state to avoid inductor coupling.  
  pinMode(PULSE_IN_A , OUTPUT);
  digitalWrite(PULSE_IN_A, LOW); 

  pinMode(PULSE_IN_B , OUTPUT);
  digitalWrite(PULSE_IN_B, LOW); 

  pinMode(EN_VREF , OUTPUT);
  digitalWrite(EN_VREF, LOW); 
  
  pinMode(AIN0, INPUT);

  pinMode(LED , OUTPUT);
  digitalWrite(LED, LOW);  

  //get index from eeprom
  uint32_t bufferRotationIndex;
  for(uint8_t i = 0; i < B_SIZE; ++i){
    EEPROM.get(B_ADDRESS + 4*i, bufferRotationIndex);
    if(bufferRotationIndex > rotationIndex){
      rotationIndex = bufferRotationIndex;
      currentBufferPosition = i;
    }
  }
  

  attachInterrupt(digitalPinToInterrupt(2), requestCalibration, FALLING);
  
  cli();
  ADCSRA &= ~(1<<ADEN);               // disable ADC
  ADCSRB = (1 << ACME);               // Enable multiplexer (PCx as AIN1)
  DIDR1 = (1<<AIN1D) | (1<< AIN0D);   // Disable comparator input buffer
  sei();
  
  delay(100);
  

  LoRa.setPins(LORA_SS, LORA_RST, LORA_DIO0);
  while (!LoRa.begin(433E6)) {
    Serial.println(".");
    delay(500);
  } 

  LoRa.setSpreadingFactor(8);
  LoRa.setSignalBandwidth(500000);
  LoRa.setCodingRate4(5);
  LoRa.setTxPower(0);
  LoRa.sleep();

  //read values from EEPROM
  pulseCntAMax = EEPROM.read(0x00);
  pulseCntAMin = EEPROM.read(0x01);
  pulseCntBMax = EEPROM.read(0x02);
  pulseCntBMin = EEPROM.read(0x03);
  thresholdA = EEPROM.read(0x04);
  thresholdB = EEPROM.read(0x05);

  
  #if LOGDEBUG==1
  Serial.println("Read from EEPROM:");
  Serial.println(String("pulseCntAMax: ") + pulseCntAMax);
  Serial.println(String("pulseCntAMin: ") + pulseCntAMin);
  Serial.println(String("pulseCntBMax: ") + pulseCntBMax);
  Serial.println(String("pulseCntBMin: ") + pulseCntBMin);
  Serial.println(String("thresholdA: ") + thresholdA);
  Serial.println(String("thresholdB: ") + thresholdB);
  #endif

}

void loop() {
  if(calibrationRequested){
    runCalibration();
    calibrationRequested = false;
  }

  // Enable 2.5V for bias and reference
  unsigned long t1 = micros();
  digitalWrite(EN_VREF, HIGH);
  _delay_us(200);
    
  pulseCntA = countPulseA();
  LowPower.powerDown(SLEEP_15MS, ADC_ON, BOD_OFF);
  pulseCntB = countPulseB();
 
  digitalWrite(EN_VREF, LOW);
  unsigned long t2 = micros()-t1;
  Serial.println(t2);

  if(updateState()){
    slowMode = false;
    lastChangeTime = millis();
    Serial.println("SlowMode disabled");
  } else {
    //If the status has not changed for 10s, we switch to slow mode with more deep sleep
    //Millis is not updated during power down sleep
    if(!slowMode && (millis() - lastChangeTime > 500)){
      slowMode = true;
      Serial.println("SlowMode enabled");
    }
  }


  #if LOGDEBUG==1
  Serial.println(String("A:") + pulseCntA + ", B:" + pulseCntB + ", rotationState:" + String(rotationState, BIN) + ", index:" + rotationIndex + ", error:" + errorIndex);
  delay(10);
  #endif

  ++measureCounter;
  if(measureCounter >= SEND_PERIOD){
    measureCounter = 0;
    transmitData();
  }
  
  // Save battery, keep ADC_ON param to avoid to enable it after power down. ADC is already disabled for the analog comparator.
  LowPower.powerDown(slowMode?SLEEP_1S:SLEEP_250MS, ADC_ON, BOD_OFF);
}

uint8_t countPulseA(){  
  pinMode(PULSE_IN_A , INPUT);

  // Set PC0 or PC1 as remplacement for AIN1
  ADMUX = PULSE_IN_A-14;

  pulseCntInt = 0;

  // Start oscillation 
  PORTD |= 0b00001000;
  
  //wait
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");

  
  // Pulse out as input to rise up impedance and not disturb oscillation and stop  
  PORTD &= 0b11110111;
  DDRD &= 0b11110111;
  


  // Enable analog comparator interrupts on rising and falling edge, enable the comparator
  ACSR = (1<<ACIE);              
  _delay_us(200);
  
  // Disable comparator and its interrupts
  ACSR = (1<<ACD);   

  // Pulse in/out as output with low state to avoid inductor coupling. 
  pinMode(PULSE_IN_A , OUTPUT);
  digitalWrite(PULSE_IN_A, LOW); 

  pinMode(PULSE_OUT_A , OUTPUT);
  digitalWrite(PULSE_OUT_A, LOW); 
               
  return pulseCntInt;    
}


uint8_t countPulseB(){  
  pinMode(PULSE_IN_B , INPUT);
  
  // Set PC0 or PC1 as remplacement for AIN1
  ADMUX = PULSE_IN_B-14;

  pulseCntInt = 0;

  // Start oscillation 
  PORTD |= 0b00010000;
  
  //wait
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  __asm__ __volatile__("nop");
  
  // Pulse out as input to rise up impedance and not disturb oscillation and stop
  PORTD &= 0b11101111;
  DDRD &=  0b11101111;     


  // Enable analog comparator interrupts on rising and falling edge, enable the comparator
  ACSR = (1<<ACIE);              
  _delay_us(200);
  
  // Disable comparator and its interrupts
  ACSR = (1<<ACD);   

  // Pulse in/out as output with low state to avoid inductor coupling. 
  pinMode(PULSE_IN_B , OUTPUT);
  digitalWrite(PULSE_IN_B, LOW); 

  pinMode(PULSE_OUT_B , OUTPUT);
  digitalWrite(PULSE_OUT_B, LOW); 
               
  return pulseCntInt;    
}

void updateEEPROM(){
  ++currentBufferPosition;
  if(currentBufferPosition == B_SIZE){
    currentBufferPosition = 0;
  }
  EEPROM.put(B_ADDRESS + 4*currentBufferPosition, rotationIndex);
}

void transmitData(){
  /* Here write your own data packet 
  * with plainRadioBuffer
  * with transfertRadioBuffer
  * rotationIndex
  
  * pulseCntA
  * pulseCntAMax
  * pulseCntAMin
  * thresholdA

  * pulseCntB
  * pulseCntBMax
  * pulseCntBMin
  * thresholdB

  * rotationState

*/
  
  LoRa.beginPacket();
  LoRa.write(transfertRadioBuffer, 30);
  LoRa.endPacket();
  LoRa.sleep();
}

void runCalibration(){
  Serial.println("Start calibration...");
  
  // Enable 2.5V for bias and reference
  digitalWrite(EN_VREF, HIGH);
  delayMicroseconds(200);

  pulseCntAMax = countPulseA();
  delayMicroseconds(100);
  pulseCntBMax = countPulseB();

  pulseCntAMin = pulseCntAMax;
  pulseCntBMin = pulseCntBMax;  
    
  for(uint8_t i = 0; i < 255; ++i){
    digitalWrite(LED, HIGH);
  
    pulseCntA = countPulseA();
    delayMicroseconds(100);
    pulseCntB = countPulseB();
    
    if(pulseCntA > pulseCntAMax) pulseCntAMax = pulseCntA;
    if(pulseCntB > pulseCntBMax) pulseCntBMax = pulseCntB;

    if(pulseCntA < pulseCntAMin) pulseCntAMin = pulseCntA;
    if(pulseCntB < pulseCntBMin) pulseCntBMin = pulseCntB;

    delay(125);
    digitalWrite(LED, LOW);
    delay(125);
  }

  thresholdA = uint16_t(pulseCntAMax + pulseCntAMin + pulseCntAMin) / 3;
  thresholdB = uint16_t(pulseCntBMax + pulseCntBMin + pulseCntBMin) / 3;

  Serial.println("End calibration:");
  Serial.println(String("pulseCntAMax: ") + pulseCntAMax);
  Serial.println(String("pulseCntAMin: ") + pulseCntAMin);
  Serial.println(String("pulseCntBMax: ") + pulseCntBMax);
  Serial.println(String("pulseCntBMin: ") + pulseCntBMin);
  Serial.println(String("thresholdA: ") + thresholdA);
  Serial.println(String("thresholdB: ") + thresholdB);

  EEPROM.write(0x00, pulseCntAMax);
  EEPROM.write(0x01, pulseCntAMin);
  EEPROM.write(0x02, pulseCntBMax);
  EEPROM.write(0x03, pulseCntBMin);
  EEPROM.write(0x04, thresholdA);
  EEPROM.write(0x05, thresholdB);

  
  digitalWrite(EN_VREF, LOW);
  measureCounter = 0;
  transmitData();
}


// Comparator interrupt
ISR(ANALOG_COMP_vect){
  pulseCntInt++;
}

void requestCalibration(){
    calibrationRequested = true;
}
